@@ -56,11 +56,21 @@ module Arg = struct
 
   let pos = pos
 
+  let pos_all = pos_all
+
   let file = file
 
+  let dir = dir
+
   let bool = bool
 
   let ( & ) = ( & )
+
+  let conv = conv
+  let conv_parser = conv_parser
+  let conv_printer = conv_printer
+
+  let non_empty = non_empty
 end
 
 [@@@ocaml.warning "+3"]
