@@ -21,32 +21,48 @@ open Types
 module O = Codefmt
 open O.Infix
 
+let tag tag t = O.span ~attr:tag t
+
+let rec filter_map acc f = function
+  | hd :: tl ->
+      let acc = match f hd with Some x -> x :: acc | None -> acc in
+      filter_map acc f tl
+  | [] -> List.rev acc
+
 let label t =
   match t with
-  | Odoc_model.Lang.TypeExpr.Label s -> O.txt s
-  | Optional s -> O.txt "?" ++ O.txt s
-
-let tag tag t = O.span ~attr:tag t
+  | Odoc_model.Lang.TypeExpr.Label s -> tag "label" (O.txt s)
+  | Optional s -> tag "optlabel" (O.txt "?" ++ O.txt s)
 
 let type_var tv = tag "type-var" (O.txt tv)
 
 let enclose ~l ~r x = O.span (O.txt l ++ x ++ O.txt r)
 
-let path p txt =
-  O.elt
-    [ inline @@ InternalLink (InternalLink.Resolved (Url.from_path p, txt)) ]
+let resolved p content =
+  let link = { InternalLink.target = Resolved p; content; tooltip = None } in
+  O.elt [ inline @@ InternalLink link ]
 
-let resolved p txt =
-  O.elt [ inline @@ InternalLink (InternalLink.Resolved (p, txt)) ]
+let path p content = resolved (Url.from_path p) content
 
-let unresolved txt =
-  O.elt [ inline @@ InternalLink (InternalLink.Unresolved txt) ]
+let unresolved content =
+  let link = { InternalLink.target = Unresolved; content; tooltip = None } in
+  O.elt [ inline @@ InternalLink link ]
 
 let path_to_id path =
   match Url.Anchor.from_identifier (path :> Paths.Identifier.t) with
   | Error _ -> None
   | Ok url -> Some url
 
+let source_anchor locs =
+  (* Remove when dropping support for OCaml < 4.08 *)
+  let to_option = function Result.Ok x -> Some x | Result.Error _ -> None in
+  match locs with
+  | Some id ->
+      Url.Anchor.from_identifier
+        (id : Paths.Identifier.SourceLocation.t :> Paths.Identifier.t)
+      |> to_option
+  | _ -> None
+
 let attach_expansion ?(status = `Default) (eq, o, e) page text =
   match page with
   | None -> O.documentedSrc text
@@ -61,6 +77,10 @@ let attach_expansion ?(status = `Default) (eq, o, e) page text =
       DocumentedSrc.
         [ Alternative (Expansion { summary; url; status; expansion }) ]
 
+let mk_heading ?(level = 1) ?label text =
+  let title = [ inline @@ Text text ] in
+  Item.Heading { label; level; title; source_anchor = None }
+
 (** Returns the preamble as an item. Stop the preamble at the first heading. The
     rest is inserted into [items]. *)
 let prepare_preamble comment items =
@@ -72,10 +92,10 @@ let prepare_preamble comment items =
   in
   (Comment.standalone preamble, Comment.standalone first_comment @ items)
 
-let make_expansion_page url comments items =
+let make_expansion_page ~source_anchor url comments items =
   let comment = List.concat comments in
   let preamble, items = prepare_preamble comment items in
-  { Page.preamble; items; url }
+  { Page.preamble; items; url; source_anchor }
 
 include Generator_signatures
 
@@ -103,7 +123,6 @@ module Make (Syntax : SYNTAX) = struct
           link1 ++ O.txt "(" ++ link2 ++ O.txt ")"
       | `Resolved _ when Paths.Path.is_hidden path ->
           let txt = Url.render_path path in
-          Format.eprintf "Warning, resolved hidden path: %s\n%!" txt;
           unresolved [ inline @@ Text txt ]
       | `Resolved rp -> (
           (* If the path is pointing to an opaque module or module type
@@ -183,6 +202,82 @@ module Make (Syntax : SYNTAX) = struct
           unresolved [ inline @@ Text txt ]
   end
 
+  module Impl = struct
+    let impl ~infos src =
+      let l =
+        infos
+        |> List.sort (fun (_, (l1, e1)) (_, (l2, e2)) ->
+               if l1 = l2 then compare e2 e1
+                 (* If two intervals open at the same time, we open
+                    first the one that closes last *)
+               else compare l1 l2)
+      in
+      let get_src a b =
+        let in_bound x = min (max x 0) (String.length src) in
+        let a = in_bound a and b = in_bound b in
+        let a, b = (min a b, max a b) in
+        String.sub src a (b - a)
+      in
+      let plain_code = function
+        | "" -> []
+        | s -> [ Types.Source_page.Plain_code s ]
+      in
+      let min (a : int) b = if a < b then a else b in
+      let rec extract from to_ list aux =
+        match list with
+        | (k, (loc_start, loc_end)) :: q when loc_start < to_ ->
+            let loc_end = min loc_end to_ in
+            (* In case of inconsistent [a  [b    a] b]
+               we do                   [a  [b  b]a] *)
+            let initial = plain_code (get_src from loc_start) in
+            let next, q = extract loc_start loc_end q [] in
+            extract loc_end to_ q
+              ([ Types.Source_page.Tagged_code (k, List.rev next) ]
+              @ initial @ aux)
+        | q -> (plain_code (get_src from to_) @ aux, q)
+      in
+      let doc, _ = extract 0 (String.length src) l [] in
+      List.rev doc
+  end
+
+  module Source_page : sig
+    val url : Paths.Identifier.SourcePage.t -> Url.t
+    val source :
+      Paths.Identifier.SourcePage.t ->
+      Syntax_highlighter.infos ->
+      Lang.Source_info.infos ->
+      string ->
+      Source_page.t
+  end = struct
+    let path id = Url.Path.from_identifier id
+    let url id = Url.from_path (path id)
+
+    let info_of_info : Lang.Source_info.annotation -> Source_page.info option =
+      function
+      | Value id -> (
+          match Url.Anchor.from_identifier (id :> Paths.Identifier.t) with
+          | Ok url -> Some (Link url)
+          | Error _ -> None)
+      | Definition id -> (
+          match id.iv with
+          | `SourceLocation (_, def) -> Some (Anchor (DefName.to_string def))
+          | `SourceLocationInternal (_, local) ->
+              Some (Anchor (LocalName.to_string local))
+          | _ -> None)
+
+    let source id syntax_info infos source_code =
+      let url = path id in
+      let mapper (info, loc) =
+        match info_of_info info with Some x -> Some (x, loc) | None -> None
+      in
+      let infos = filter_map [] mapper infos in
+      let syntax_info =
+        List.map (fun (ty, loc) -> (Source_page.Syntax ty, loc)) syntax_info
+      in
+      let contents = Impl.impl ~infos:(infos @ syntax_info) source_code in
+      { Source_page.url; contents }
+  end
+
   module Type_expression : sig
     val type_expr : ?needs_parentheses:bool -> Lang.TypeExpr.t -> text
 
@@ -229,8 +324,8 @@ module Make (Syntax : SYNTAX) = struct
                           let arguments = style_arguments ~constant arguments in
                           O.span
                             (if Syntax.Type.Variant.parenthesize_params then
-                             constr ++ arguments
-                            else constr ++ O.txt " of" ++ O.sp ++ arguments))
+                               constr ++ arguments
+                             else constr ++ O.txt " of" ++ O.sp ++ arguments))
                   in
                   if add_pipe then O.sp ++ res else res
             in
@@ -375,6 +470,8 @@ module Make (Syntax : SYNTAX) = struct
 
     val extension : Lang.Extension.t -> Item.t
 
+    val record : Lang.TypeDecl.Field.t list -> DocumentedSrc.one list
+
     val exn : Lang.Exception.t -> Item.t
 
     val format_params :
@@ -460,11 +557,11 @@ module Make (Syntax : SYNTAX) = struct
           O.documentedSrc
             (cstr
             ++ (if Syntax.Type.Variant.parenthesize_params then
-                O.txt "(" ++ params ++ O.txt ")"
-               else
-                 (if is_gadt then O.txt Syntax.Type.annotation_separator
-                 else O.txt " " ++ O.keyword "of" ++ O.txt " ")
-                 ++ params)
+                  O.txt "(" ++ params ++ O.txt ")"
+                else
+                  (if is_gadt then O.txt Syntax.Type.annotation_separator
+                   else O.txt " " ++ O.keyword "of" ++ O.txt " ")
+                  ++ params)
             ++ ret_type)
       | Record fields ->
           if is_gadt then
@@ -538,7 +635,9 @@ module Make (Syntax : SYNTAX) = struct
       | Error e -> failwith (Url.Error.to_string e)
       | Ok url ->
           let anchor = Some url in
-          let attrs = [ "def"; Url.Anchor.string_of_kind url.kind ] in
+          let attrs =
+            [ "def"; "variant"; Url.Anchor.string_of_kind url.kind ]
+          in
           let code =
             O.documentedSrc (O.txt "| ") @ constructor id t.args t.res
           in
@@ -551,6 +650,9 @@ module Make (Syntax : SYNTAX) = struct
         O.keyword "type" ++ O.txt " "
         ++ Link.from_path (t.type_path :> Paths.Path.t)
         ++ O.txt " +=" ++ O.sp
+        ++
+        if t.private_ then O.keyword Syntax.Type.private_keyword ++ O.sp
+        else O.noop
       in
       let content =
         O.documentedSrc prefix
@@ -561,7 +663,14 @@ module Make (Syntax : SYNTAX) = struct
       let attr = [ "type"; "extension" ] in
       let anchor = Some (Url.Anchor.extension_decl t) in
       let doc = Comment.to_ir t.doc in
-      Item.Declaration { attr; anchor; doc; content }
+      let source_anchor =
+        (* Take the anchor from the first constructor only for consistency with
+           regular variants. *)
+        match t.constructors with
+        | hd :: _ -> source_anchor hd.locs
+        | [] -> None
+      in
+      Item.Declaration { attr; anchor; doc; content; source_anchor }
 
     let exn (t : Odoc_model.Lang.Exception.t) =
       let cstr = constructor (t.id :> Paths.Identifier.t) t.args t.res in
@@ -574,7 +683,8 @@ module Make (Syntax : SYNTAX) = struct
       let attr = [ "exception" ] in
       let anchor = path_to_id t.id in
       let doc = Comment.to_ir t.doc in
-      Item.Declaration { attr; anchor; doc; content }
+      let source_anchor = source_anchor t.locs in
+      Item.Declaration { attr; anchor; doc; content; source_anchor }
 
     let polymorphic_variant ~type_ident
         (t : Odoc_model.Lang.TypeExpr.Polymorphic_variant.t) =
@@ -582,12 +692,12 @@ module Make (Syntax : SYNTAX) = struct
         let kind_approx, cstr, doc =
           match item with
           | Odoc_model.Lang.TypeExpr.Polymorphic_variant.Type te ->
-              ("unknown", O.code (type_expr te), None)
+              ("unknown", O.documentedSrc (type_expr te), None)
           | Constructor { constant; name; arguments; doc; _ } -> (
               let cstr = "`" ^ name in
               ( "constructor",
                 (match arguments with
-                | [] -> O.code (O.txt cstr)
+                | [] -> O.documentedSrc (O.txt cstr)
                 | _ ->
                     (* Multiple arguments in a polymorphic variant constructor correspond
                        to a conjunction of types, not a product: [`Lbl int&float].
@@ -609,7 +719,7 @@ module Make (Syntax : SYNTAX) = struct
                     let params =
                       if constant then O.txt "& " ++ params else params
                     in
-                    O.code
+                    O.documentedSrc
                       (O.txt cstr
                       ++
                       if Syntax.Type.Variant.parenthesize_params then params
@@ -619,18 +729,20 @@ module Make (Syntax : SYNTAX) = struct
         let markers = Syntax.Comment.markers in
         try
           let url = Url.Anchor.polymorphic_variant ~type_ident item in
-          let attrs = [ "def"; Url.Anchor.string_of_kind url.kind ] in
+          let attrs =
+            [ "def"; "variant"; Url.Anchor.string_of_kind url.kind ]
+          in
           let anchor = Some url in
-          let code = O.code (O.txt "| ") @ cstr in
+          let code = O.documentedSrc (O.txt "| ") @ cstr in
           let doc = match doc with None -> [] | Some doc -> doc in
-          DocumentedSrc.Documented { attrs; anchor; code; doc; markers }
+          DocumentedSrc.Nested { attrs; anchor; code; doc; markers }
         with Failure s ->
           Printf.eprintf "ERROR: %s\n%!" s;
-          let code = O.code (O.txt "| ") @ cstr in
+          let code = O.documentedSrc (O.txt "| ") @ cstr in
           let attrs = [ "def"; kind_approx ] in
           let doc = [] in
           let anchor = None in
-          DocumentedSrc.Documented { attrs; anchor; code; doc; markers }
+          DocumentedSrc.Nested { attrs; anchor; code; doc; markers }
       in
       let variants = List.map row t.elements in
       let intro, ending =
@@ -704,8 +816,8 @@ module Make (Syntax : SYNTAX) = struct
             O.txt (if is_substitution then " :=" else " =")
             ++ O.sp
             ++ (if private_ then
-                O.keyword Syntax.Type.private_keyword ++ O.txt " "
-               else O.noop)
+                  O.keyword Syntax.Type.private_keyword ++ O.txt " "
+                else O.noop)
             ++ type_expr t
           in
           (manifest, false)
@@ -784,7 +896,8 @@ module Make (Syntax : SYNTAX) = struct
       let attr = "type" :: (if is_substitution then [ "subst" ] else []) in
       let anchor = path_to_id t.id in
       let doc = Comment.to_ir t.doc in
-      Item.Declaration { attr; anchor; doc; content }
+      let source_anchor = source_anchor t.locs in
+      Item.Declaration { attr; anchor; doc; content; source_anchor }
   end
 
   open Type_declaration
@@ -798,6 +911,7 @@ module Make (Syntax : SYNTAX) = struct
         | Abstract -> ([], Syntax.Value.semicolon)
         | External _ -> ([ "external" ], Syntax.Type.External.semicolon)
       in
+      (* TODO: link to source *)
       let name = Paths.Identifier.name t.id in
       let content =
         O.documentedSrc
@@ -811,7 +925,8 @@ module Make (Syntax : SYNTAX) = struct
       let attr = [ "value" ] @ extra_attr in
       let anchor = path_to_id t.id in
       let doc = Comment.to_ir t.doc in
-      Item.Declaration { attr; anchor; doc; content }
+      let source_anchor = source_anchor t.locs in
+      Item.Declaration { attr; anchor; doc; content; source_anchor }
   end
 
   open Value
@@ -911,7 +1026,7 @@ module Make (Syntax : SYNTAX) = struct
       let attr = [ "method" ] in
       let anchor = path_to_id t.id in
       let doc = Comment.to_ir t.doc in
-      Item.Declaration { attr; anchor; doc; content }
+      Item.Declaration { attr; anchor; doc; content; source_anchor = None }
 
     let instance_variable (t : Odoc_model.Lang.InstanceVariable.t) =
       let name = Paths.Identifier.name t.id in
@@ -930,7 +1045,7 @@ module Make (Syntax : SYNTAX) = struct
       let attr = [ "value"; "instance-variable" ] in
       let anchor = path_to_id t.id in
       let doc = Comment.to_ir t.doc in
-      Item.Declaration { attr; anchor; doc; content }
+      Item.Declaration { attr; anchor; doc; content; source_anchor = None }
 
     let inherit_ (ih : Lang.ClassSignature.Inherit.t) =
       let cte =
@@ -944,7 +1059,7 @@ module Make (Syntax : SYNTAX) = struct
       let attr = [ "inherit" ] in
       let anchor = None in
       let doc = Comment.to_ir ih.doc in
-      Item.Declaration { attr; anchor; doc; content }
+      Item.Declaration { attr; anchor; doc; content; source_anchor = None }
 
     let constraint_ (cst : Lang.ClassSignature.Constraint.t) =
       let content =
@@ -953,7 +1068,7 @@ module Make (Syntax : SYNTAX) = struct
       let attr = [] in
       let anchor = None in
       let doc = Comment.to_ir cst.doc in
-      Item.Declaration { attr; anchor; doc; content }
+      Item.Declaration { attr; anchor; doc; content; source_anchor = None }
 
     let class_signature (c : Lang.ClassSignature.t) =
       let rec loop l acc_items =
@@ -998,18 +1113,26 @@ module Make (Syntax : SYNTAX) = struct
 
     let class_ (t : Odoc_model.Lang.Class.t) =
       let name = Paths.Identifier.name t.id in
-      let params = format_params ~delim:`brackets t.params in
+      let params =
+        match t.params with
+        | [] -> O.noop
+        | _ :: _ as params -> format_params ~delim:`brackets params ++ O.txt " "
+      in
       let virtual_ =
         if t.virtual_ then O.keyword "virtual" ++ O.txt " " else O.noop
       in
 
+      let source_anchor = source_anchor t.locs in
       let cname, expansion, expansion_doc =
         match t.expansion with
         | None -> (O.documentedSrc @@ O.txt name, None, None)
         | Some csig ->
             let expansion_doc, items = class_signature csig in
             let url = Url.Path.from_identifier t.id in
-            let page = make_expansion_page url [ t.doc; expansion_doc ] items in
+            let page =
+              make_expansion_page ~source_anchor url [ t.doc; expansion_doc ]
+                items
+            in
             ( O.documentedSrc @@ path url [ inline @@ Text name ],
               Some page,
               Some expansion_doc )
@@ -1023,14 +1146,13 @@ module Make (Syntax : SYNTAX) = struct
           expansion summary
       in
       let content =
-        O.documentedSrc
-          (O.keyword "class" ++ O.txt " " ++ virtual_ ++ params ++ O.txt " ")
+        O.documentedSrc (O.keyword "class" ++ O.txt " " ++ virtual_ ++ params)
         @ cname @ cd
       in
       let attr = [ "class" ] in
       let anchor = path_to_id t.id in
       let doc = Comment.synopsis ~decl_doc:t.doc ~expansion_doc in
-      Item.Declaration { attr; anchor; doc; content }
+      Item.Declaration { attr; anchor; doc; content; source_anchor }
 
     let class_type (t : Odoc_model.Lang.ClassType.t) =
       let name = Paths.Identifier.name t.id in
@@ -1038,13 +1160,17 @@ module Make (Syntax : SYNTAX) = struct
       let virtual_ =
         if t.virtual_ then O.keyword "virtual" ++ O.txt " " else O.noop
       in
+      let source_anchor = source_anchor t.locs in
       let cname, expansion, expansion_doc =
         match t.expansion with
         | None -> (O.documentedSrc @@ O.txt name, None, None)
         | Some csig ->
             let url = Url.Path.from_identifier t.id in
             let expansion_doc, items = class_signature csig in
-            let page = make_expansion_page url [ t.doc; expansion_doc ] items in
+            let page =
+              make_expansion_page ~source_anchor url [ t.doc; expansion_doc ]
+                items
+            in
             ( O.documentedSrc @@ path url [ inline @@ Text name ],
               Some page,
               Some expansion_doc )
@@ -1060,7 +1186,7 @@ module Make (Syntax : SYNTAX) = struct
       let attr = [ "class-type" ] in
       let anchor = path_to_id t.id in
       let doc = Comment.synopsis ~decl_doc:t.doc ~expansion_doc in
-      Item.Declaration { attr; anchor; doc; content }
+      Item.Declaration { attr; anchor; doc; content; source_anchor }
   end
 
   open Class
@@ -1168,7 +1294,10 @@ module Make (Syntax : SYNTAX) = struct
             let url = Url.Path.from_identifier arg.id in
             let modname = path url [ inline @@ Text name ] in
             let type_with_expansion =
-              let content = make_expansion_page url [ expansion_doc ] items in
+              let content =
+                make_expansion_page ~source_anchor:None url [ expansion_doc ]
+                  items
+              in
               let summary = O.render modtyp in
               let status = `Default in
               let expansion =
@@ -1199,17 +1328,18 @@ module Make (Syntax : SYNTAX) = struct
       let attr = [ "module-substitution" ] in
       let anchor = path_to_id t.id in
       let doc = Comment.to_ir t.doc in
-      Item.Declaration { attr; anchor; doc; content }
+      Item.Declaration { attr; anchor; doc; content; source_anchor = None }
 
     and module_type_substitution (t : Odoc_model.Lang.ModuleTypeSubstitution.t)
         =
       let prefix =
         O.keyword "module" ++ O.txt " " ++ O.keyword "type" ++ O.txt " "
       in
+      let source_anchor = None in
       let modname = Paths.Identifier.name t.id in
       let modname, expansion_doc, mty =
-        module_type_manifest ~subst:true modname t.id t.doc (Some t.manifest)
-          prefix
+        module_type_manifest ~subst:true ~source_anchor modname t.id t.doc
+          (Some t.manifest) prefix
       in
       let content =
         O.documentedSrc (prefix ++ modname)
@@ -1220,7 +1350,7 @@ module Make (Syntax : SYNTAX) = struct
       let attr = [ "module-type" ] in
       let anchor = path_to_id t.id in
       let doc = Comment.synopsis ~decl_doc:t.doc ~expansion_doc in
-      Item.Declaration { attr; anchor; doc; content }
+      Item.Declaration { attr; anchor; doc; content; source_anchor }
 
     and simple_expansion :
         Odoc_model.Lang.ModuleType.simple_expansion ->
@@ -1251,25 +1381,13 @@ module Make (Syntax : SYNTAX) = struct
                   @@ Url.Anchor.from_identifier (arg.id :> Paths.Identifier.t)
                 in
                 let doc = [] in
-                [ Item.Declaration { content; anchor; attr; doc } ])
-          in
-          let prelude =
-            Item.Heading
-              {
-                label = Some "parameters";
-                level = 1;
-                title = [ inline @@ Text "Parameters" ];
-              }
-            :: params
-          and content =
-            Item.Heading
-              {
-                label = Some "signature";
-                level = 1;
-                title = [ inline @@ Text "Signature" ];
-              }
-            :: content
+                [
+                  Item.Declaration
+                    { content; anchor; attr; doc; source_anchor = None };
+                ])
           in
+          let prelude = mk_heading ~label:"parameters" "Parameters" :: params
+          and content = mk_heading ~label:"signature" "Signature" :: content in
           (sg_doc, prelude @ content)
 
     and expansion_of_module_type_expr :
@@ -1305,6 +1423,7 @@ module Make (Syntax : SYNTAX) = struct
         | Alias (_, None) -> None
         | ModuleType e -> expansion_of_module_type_expr e
       in
+      let source_anchor = source_anchor t.locs in
       let modname, status, expansion, expansion_doc =
         match expansion with
         | None -> (O.txt modname, `Default, None, None)
@@ -1316,9 +1435,13 @@ module Make (Syntax : SYNTAX) = struct
             in
             let url = Url.Path.from_identifier t.id in
             let link = path url [ inline @@ Text modname ] in
-            let page = make_expansion_page url [ t.doc; expansion_doc ] items in
+            let page =
+              make_expansion_page ~source_anchor url [ t.doc; expansion_doc ]
+                items
+            in
             (link, status, Some page, Some expansion_doc)
       in
+      (* TODO: link to source *)
       let intro = O.keyword "module" ++ O.txt " " ++ modname in
       let summary = O.ignore intro ++ mdexpr_in_decl t.id t.type_ in
       let modexpr =
@@ -1334,7 +1457,7 @@ module Make (Syntax : SYNTAX) = struct
       let attr = [ "module" ] in
       let anchor = path_to_id t.id in
       let doc = Comment.synopsis ~decl_doc:t.doc ~expansion_doc in
-      Item.Declaration { attr; anchor; doc; content }
+      Item.Declaration { attr; anchor; doc; content; source_anchor }
 
     and simple_expansion_in_decl (base : Paths.Identifier.Module.t) se =
       let rec ty_of_se :
@@ -1360,7 +1483,8 @@ module Make (Syntax : SYNTAX) = struct
       | Alias (mod_path, _) -> Link.from_path (mod_path :> Paths.Path.t)
       | ModuleType mt -> mty mt
 
-    and module_type_manifest ~subst modname id doc manifest prefix =
+    and module_type_manifest ~subst ~source_anchor modname id doc manifest
+        prefix =
       let expansion =
         match manifest with
         | None -> None
@@ -1372,7 +1496,10 @@ module Make (Syntax : SYNTAX) = struct
         | Some (expansion_doc, items) ->
             let url = Url.Path.from_identifier id in
             let link = path url [ inline @@ Text modname ] in
-            let page = make_expansion_page url [ doc; expansion_doc ] items in
+            let page =
+              make_expansion_page ~source_anchor url [ doc; expansion_doc ]
+                items
+            in
             (link, Some page, Some expansion_doc)
       in
       let summary =
@@ -1392,8 +1519,10 @@ module Make (Syntax : SYNTAX) = struct
         O.keyword "module" ++ O.txt " " ++ O.keyword "type" ++ O.txt " "
       in
       let modname = Paths.Identifier.name t.id in
+      let source_anchor = source_anchor t.locs in
       let modname, expansion_doc, mty =
-        module_type_manifest ~subst:false modname t.id t.doc t.expr prefix
+        module_type_manifest ~subst:false ~source_anchor modname t.id t.doc
+          t.expr prefix
       in
       let content =
         O.documentedSrc (prefix ++ modname)
@@ -1404,7 +1533,7 @@ module Make (Syntax : SYNTAX) = struct
       let attr = [ "module-type" ] in
       let anchor = path_to_id t.id in
       let doc = Comment.synopsis ~decl_doc:t.doc ~expansion_doc in
-      Item.Declaration { attr; anchor; doc; content }
+      Item.Declaration { attr; anchor; doc; content; source_anchor }
 
     and umty_hidden : Odoc_model.Lang.ModuleType.U.expr -> bool = function
       | Path p -> Paths.Path.(is_hidden (p :> t))
@@ -1613,21 +1742,22 @@ module Make (Syntax : SYNTAX) = struct
            The documentation from the expansion is not used. *)
         Comment.to_ir t.doc
       in
-      Item.Include { attr; anchor; doc; content }
+      Item.Include { attr; anchor; doc; content; source_anchor = None }
   end
 
   open Module
 
   module Page : sig
-    val compilation_unit : Lang.Compilation_unit.t -> Page.t
+    val compilation_unit : Lang.Compilation_unit.t -> Document.t
 
-    val page : Lang.Page.t -> Page.t
+    val page : Lang.Page.t -> Document.t
+
+    val source_tree : Lang.SourceTree.t -> Document.t list
   end = struct
-    let pack : Odoc_model.Lang.Compilation_unit.Packed.t -> Item.t list =
+    let pack : Lang.Compilation_unit.Packed.t -> Item.t list =
      fun t ->
-      let open Odoc_model.Lang in
       let f x =
-        let id = x.Compilation_unit.Packed.id in
+        let id = x.Lang.Compilation_unit.Packed.id in
         let modname = Paths.Identifier.name id in
         let md_def =
           O.keyword "module" ++ O.txt " " ++ O.txt modname ++ O.txt " = "
@@ -1640,29 +1770,130 @@ module Make (Syntax : SYNTAX) = struct
         in
         let attr = [ "modules" ] in
         let doc = [] in
-        let decl = { Item.anchor; content; attr; doc } in
+        let decl = { Item.anchor; content; attr; doc; source_anchor = None } in
         Item.Declaration decl
       in
       List.map f t
 
-    let compilation_unit (t : Odoc_model.Lang.Compilation_unit.t) : Page.t =
+    let compilation_unit (t : Odoc_model.Lang.Compilation_unit.t) =
       let url = Url.Path.from_identifier t.id in
       let unit_doc, items =
         match t.content with
         | Module sign -> signature sign
         | Pack packed -> ([], pack packed)
       in
-      make_expansion_page url [ unit_doc ] items
+      let source_anchor =
+        match t.source_info with
+        | Some src -> Some (Source_page.url src.id)
+        | None -> None
+      in
+      let page = make_expansion_page ~source_anchor url [ unit_doc ] items in
+      Document.Page page
 
-    let page (t : Odoc_model.Lang.Page.t) : Page.t =
+    let page (t : Odoc_model.Lang.Page.t) =
       (*let name =
           match t.name.iv with `Page (_, name) | `LeafPage (_, name) -> name
         in*)
       (*let title = Odoc_model.Names.PageName.to_string name in*)
       let url = Url.Path.from_identifier t.name in
       let preamble, items = Sectioning.docs t.content in
-      { Page.preamble; items; url }
+      let source_anchor = None in
+      Document.Page { Page.preamble; items; url; source_anchor }
+
+    let source_tree t =
+      let dir_pages = t.Odoc_model.Lang.SourceTree.source_children in
+      let open Paths.Identifier in
+      let module Set = Set.Make (SourceDir) in
+      let module M = Map.Make (SourceDir) in
+      (* mmap is a from a [SourceDir.t] to its [SourceDir.t] and [SourcePage.t]
+         children *)
+      let mmap =
+        let add parent f mmap =
+          let old_value =
+            try M.find parent mmap with Not_found -> (Set.empty, [])
+          in
+          M.add parent (f old_value) mmap
+        and add_file file (set, lp) = (set, file :: lp)
+        and add_dir dir (set, lp) = (Set.add dir set, lp) in
+        let rec dir_ancestors_add dir mmap =
+          match dir.iv with
+          | `SourceDir (parent, _) ->
+              let mmap = add parent (add_dir dir) mmap in
+              dir_ancestors_add parent mmap
+          | `Page _ -> mmap
+        in
+        let file_ancestors_add ({ iv = `SourcePage (parent, _); _ } as file)
+            mmap =
+          let mmap = add parent (add_file file) mmap in
+          dir_ancestors_add parent mmap
+        in
+        List.fold_left
+          (fun mmap file -> file_ancestors_add file mmap)
+          M.empty dir_pages
+      in
+      let page_of_dir (dir : SourceDir.t) (dir_children, file_children) =
+        let url = Url.Path.from_identifier dir in
+        let block ?(attr = []) desc = Block.{ attr; desc } in
+        let inline ?(attr = []) desc = Inline.[ { attr; desc } ] in
+        let header =
+          let title = inline (Text (name dir)) in
+          Item.Heading
+            Heading.{ label = None; level = 0; title; source_anchor = None }
+        in
+        let li ?(attr = []) name url =
+          let link url desc =
+            let content = [ Inline.{ attr = []; desc } ] and tooltip = None in
+            Inline.InternalLink
+              { InternalLink.target = Resolved url; content; tooltip }
+          in
+          [ block ~attr @@ Block.Inline (inline @@ link url (Text name)) ]
+        in
+        let li_of_child child =
+          match child with
+          | { iv = `Page _; _ } ->
+              assert false (* No [`Page] is child of a [`SourceDir] *)
+          | { iv = `SourceDir (_, name); _ } ->
+              let url = child |> Url.Path.from_identifier |> Url.from_path in
+              (name, url)
+        in
+        let li_of_file_child ({ iv = `SourcePage (_, name); _ } as child) =
+          let url = child |> Url.Path.from_identifier |> Url.from_path in
+          (name, url)
+        in
+        let items =
+          let text ?(attr = []) desc = Item.Text [ { attr; desc } ] in
+          let list l = Block.List (Block.Unordered, l) in
+          let list_of_children =
+            let dir_list =
+              Set.fold
+                (fun child acc -> li_of_child child :: acc)
+                dir_children []
+            and file_list =
+              List.map (fun child -> li_of_file_child child) file_children
+            in
+            let sort ?(attr = []) l =
+              l
+              |> List.sort (fun (n1, _) (n2, _) -> String.compare n1 n2)
+              |> List.map (fun (name, url) -> li ~attr name url)
+            in
+            sort ~attr:[ "odoc-directory" ] dir_list
+            @ sort ~attr:[ "odoc-file" ] file_list
+          in
+          header
+          :: [ text ~attr:[ "odoc-folder-list" ] @@ list list_of_children ]
+        in
+        Document.Page
+          { Types.Page.preamble = []; items; url; source_anchor = None }
+      in
+      M.fold (fun dir children acc -> page_of_dir dir children :: acc) mmap []
   end
 
   include Page
+
+  let type_expr = type_expr
+
+  let record = record
+
+  let source_page id syntax_info infos source_code =
+    Document.Source_page (Source_page.source id syntax_info infos source_code)
 end
