@@ -17,6 +17,7 @@
 open Odoc_document.Types
 module Html = Tyxml.Html
 module Doctree = Odoc_document.Doctree
+module Url = Odoc_document.Url
 
 type any = Html_types.flow5
 
@@ -31,15 +32,30 @@ type non_link_phrasing = Html_types.phrasing_without_interactive
 let mk_anchor_link id =
   [ Html.a ~a:[ Html.a_href ("#" ^ id); Html.a_class [ "anchor" ] ] [] ]
 
-let mk_anchor anchor =
+let mk_anchor config anchor =
   match anchor with
   | None -> ([], [], [])
-  | Some { Odoc_document.Url.Anchor.anchor; _ } ->
+  | _ when Config.search_result config ->
+      (* When displaying for a search result, anchor are not added as it would
+         make no sense to add them. *)
+      ([], [], [])
+  | Some { Url.Anchor.anchor; _ } ->
       let link = mk_anchor_link anchor in
       let extra_attr = [ Html.a_id anchor ] in
       let extra_class = [ "anchored" ] in
       (extra_attr, extra_class, link)
 
+let mk_link_to_source ~config ~resolve anchor =
+  match anchor with
+  | None -> []
+  | Some url ->
+      let href = Link.href ~config ~resolve url in
+      [
+        Html.a
+          ~a:[ Html.a_href href; Html.a_class [ "source_link" ] ]
+          [ Html.txt "Source" ];
+      ]
+
 let class_ (l : Class.t) = if l = [] then [] else [ Html.a_class l ]
 
 let inline_math (s : Math.t) =
@@ -80,26 +96,32 @@ and styled style ~emph_level =
   | `Superscript -> (emph_level, Html.sup ~a:[])
   | `Subscript -> (emph_level, Html.sub ~a:[])
 
-let rec internallink ~config ~emph_level ~resolve ?(a = []) (t : InternalLink.t)
-    =
-  match t with
-  | Resolved (uri, content) ->
-      let href = Link.href ~config ~resolve uri in
-      let a = (a :> Html_types.a_attrib Html.attrib list) in
-      let elt =
-        Html.a ~a:(Html.a_href href :: a) (inline_nolink ~emph_level content)
-      in
-      let elt = (elt :> phrasing Html.elt) in
-      [ elt ]
-  | Unresolved content ->
-      (* let title =
-       *   Html.a_title (Printf.sprintf "unresolved reference to %S"
-       *       (ref_to_string ref)
-       * in *)
-      let a = Html.a_class [ "xref-unresolved" ] :: a in
-      let elt = Html.span ~a (inline ~config ~emph_level ~resolve content) in
-      let elt = (elt :> phrasing Html.elt) in
-      [ elt ]
+let rec internallink ~config ~emph_level ~resolve ?(a = [])
+    { InternalLink.target; content; tooltip } =
+  let a = match tooltip with Some s -> Html.a_title s :: a | None -> a in
+  let elt =
+    match target with
+    | Resolved uri ->
+        let href = Link.href ~config ~resolve uri in
+        let content = inline_nolink ~emph_level content in
+        if Config.search_result config then
+          (* When displaying for a search result, links are displayed as regular
+             text. *)
+          Html.span ~a content
+        else
+          let a =
+            Html.a_href href :: (a :> Html_types.a_attrib Html.attrib list)
+          in
+          Html.a ~a content
+    | Unresolved ->
+        (* let title =
+         *   Html.a_title (Printf.sprintf "unresolved reference to %S"
+         *       (ref_to_string ref)
+         * in *)
+        let a = Html.a_class [ "xref-unresolved" ] :: a in
+        Html.span ~a (inline ~config ~emph_level ~resolve content)
+  in
+  [ (elt :> phrasing Html.elt) ]
 
 and inline ~config ?(emph_level = 0) ~resolve (l : Inline.t) :
     phrasing Html.elt list =
@@ -115,6 +137,11 @@ and inline ~config ?(emph_level = 0) ~resolve (l : Inline.t) :
     | Styled (style, c) ->
         let emph_level, app_style = styled style ~emph_level in
         [ app_style @@ inline ~config ~emph_level ~resolve c ]
+    | Link (_, c) when Config.search_result config ->
+        (* When displaying for a search result, links are displayed as regular
+           text. *)
+        let content = inline_nolink ~emph_level c in
+        [ Html.span ~a content ]
     | Link (href, c) ->
         let a = (a :> Html_types.a_attrib Html.attrib list) in
         let content = inline_nolink ~emph_level c in
@@ -151,10 +178,15 @@ and inline_nolink ?(emph_level = 0) (l : Inline.t) :
 let heading ~config ~resolve (h : Heading.t) =
   let a, anchor =
     match h.label with
+    | Some _ when Config.search_result config ->
+        (* When displaying for a search result, anchor are not added as it would
+           make no sense to add them. *)
+        ([], [])
     | Some id -> ([ Html.a_id id ], mk_anchor_link id)
     | None -> ([], [])
   in
   let content = inline ~config ~resolve h.title in
+  let source_link = mk_link_to_source ~config ~resolve h.source_anchor in
   let mk =
     match h.level with
     | 0 -> Html.h1
@@ -164,7 +196,15 @@ let heading ~config ~resolve (h : Heading.t) =
     | 4 -> Html.h5
     | _ -> Html.h6
   in
-  mk ~a (anchor @ content)
+  mk ~a (anchor @ content @ source_link)
+
+let text_align = function
+  | Table.Left -> [ Html.a_style "text-align:left" ]
+  | Center -> [ Html.a_style "text-align:center" ]
+  | Right -> [ Html.a_style "text-align:right" ]
+  | Default -> []
+
+let cell_kind = function `Header -> Html.th | `Data -> Html.td
 
 let rec block ~config ~resolve (l : Block.t) : flow Html.elt list =
   let as_flow x = (x : phrasing Html.elt list :> flow Html.elt list) in
@@ -181,6 +221,10 @@ let rec block ~config ~resolve (l : Block.t) : flow Html.elt list =
     | List (typ, l) ->
         let mk = match typ with Ordered -> Html.ol | Unordered -> Html.ul in
         mk_block mk (List.map (fun x -> Html.li (block ~config ~resolve x)) l)
+    | Table t ->
+        mk_block ~extra_class:[ "odoc-table" ]
+          (fun ?a x -> Html.table ?a x)
+          (mk_rows ~config ~resolve t)
     | Description l ->
         let item i =
           let a = class_ i.Description.attr in
@@ -202,6 +246,29 @@ let rec block ~config ~resolve (l : Block.t) : flow Html.elt list =
   in
   Utils.list_concat_map l ~f:one
 
+and mk_rows ~config ~resolve { align; data } =
+  let mk_row row =
+    let mk_cell ~align (x, h) =
+      let a = text_align align in
+      cell_kind ~a h (block ~config ~resolve x)
+    in
+    let alignment align =
+      match align with align :: q -> (align, q) | [] -> (Table.Default, [])
+      (* Second case is for recovering from a too short alignment list. A
+         warning should have been raised when loading the doc-comment. *)
+    in
+    let acc, _align =
+      List.fold_left
+        (fun (acc, aligns) (x, h) ->
+          let align, aligns = alignment aligns in
+          let cell = mk_cell ~align (x, h) in
+          (cell :: acc, aligns))
+        ([], align) row
+    in
+    Html.tr (List.rev acc)
+  in
+  List.map mk_row data
+
 (* This coercion is actually sound, but is not currently accepted by Tyxml.
    See https://github.com/ocsigen/tyxml/pull/265 for details
    Can be replaced by a simple type coercion once this is fixed
@@ -268,7 +335,7 @@ let rec documentedSrc ~config ~resolve (t : DocumentedSrc.t) :
                     (delim opening @ block ~config ~resolve doc @ delim closing);
                 ]
           in
-          let extra_attr, extra_class, link = mk_anchor anchor in
+          let extra_attr, extra_class, link = mk_anchor config anchor in
           let content = (content :> any Html.elt list) in
           Html.li
             ~a:(extra_attr @ class_ (attrs @ extra_class))
@@ -298,7 +365,14 @@ and items ~config ~resolve l : item Html.elt list =
         (continue_with [@tailcall]) rest content
     | Heading h :: rest ->
         (continue_with [@tailcall]) rest [ heading ~config ~resolve h ]
-    | Include { attr; anchor; doc; content = { summary; status; content } }
+    | Include
+        {
+          attr;
+          anchor;
+          source_anchor;
+          doc;
+          content = { summary; status; content };
+        }
       :: rest ->
         let doc = spec_doc_div ~config ~resolve doc in
         let included_html = (items content :> item Html.elt list) in
@@ -310,9 +384,14 @@ and items ~config ~resolve l : item Html.elt list =
           let details ~open' =
             let open' = if open' then [ Html.a_open () ] else [] in
             let summary =
-              let extra_attr, extra_class, anchor_link = mk_anchor anchor in
+              let extra_attr, extra_class, anchor_link =
+                mk_anchor config anchor
+              in
+              let link_to_source =
+                mk_link_to_source ~config ~resolve source_anchor
+              in
               let a = spec_class (attr @ extra_class) @ extra_attr in
-              Html.summary ~a @@ anchor_link
+              Html.summary ~a @@ anchor_link @ link_to_source
               @ source (inline ~config ~resolve) summary
             in
             let inner =
@@ -330,10 +409,13 @@ and items ~config ~resolve l : item Html.elt list =
           | `Default -> details ~open':(Config.open_details config)
         in
         (continue_with [@tailcall]) rest content
-    | Declaration { Item.attr; anchor; content; doc } :: rest ->
-        let extra_attr, extra_class, anchor_link = mk_anchor anchor in
+    | Declaration { Item.attr; anchor; source_anchor; content; doc } :: rest ->
+        let extra_attr, extra_class, anchor_link = mk_anchor config anchor in
+        let link_to_source = mk_link_to_source ~config ~resolve source_anchor in
         let a = spec_class (attr @ extra_class) @ extra_attr in
-        let content = anchor_link @ documentedSrc ~config ~resolve content in
+        let content =
+          anchor_link @ link_to_source @ documentedSrc ~config ~resolve content
+        in
         let spec =
           let doc = spec_doc_div ~config ~resolve doc in
           [ div ~a:[ Html.a_class [ "odoc-spec" ] ] (div ~a content :: doc) ]
@@ -402,35 +484,74 @@ module Page = struct
 
   let rec include_ ~config { Subpage.content; _ } = page ~config content
 
-  and subpages ~config subpages =
-    Utils.list_concat_map ~f:(include_ ~config) subpages
+  and subpages ~config subpages = List.map (include_ ~config) subpages
 
-  and page ~config p : Odoc_document.Renderer.page list =
-    let { Page.preamble; items = i; url } = Doctree.Labels.disambiguate_page p
-    and subpages =
-      (* Don't use the output of [disambiguate_page] to avoid unecessarily
-         mangled labels. *)
-      subpages ~config @@ Doctree.Subpages.compute p
+  and page ~config p : Odoc_document.Renderer.page =
+    let { Page.preamble; items = i; url; source_anchor } =
+      Doctree.Labels.disambiguate_page ~enter_subpages:false p
     in
+    let subpages = subpages ~config @@ Doctree.Subpages.compute p in
     let resolve = Link.Current url in
     let i = Doctree.Shift.compute ~on_sub i in
     let uses_katex = Doctree.Math.has_math_elements p in
     let toc = Toc.gen_toc ~config ~resolve ~path:url i in
     let breadcrumbs = Breadcrumbs.gen_breadcrumbs ~config ~url in
-    let header =
-      items ~config ~resolve (Doctree.PageTitle.render_title p @ preamble)
-    in
     let content = (items ~config ~resolve i :> any Html.elt list) in
     if Config.as_json config then
+      let source_anchor =
+        match source_anchor with
+        | Some url -> Some (Link.href ~config ~resolve url)
+        | None -> None
+      in
       Html_fragment_json.make ~config
         ~preamble:(items ~config ~resolve preamble :> any Html.elt list)
-        ~breadcrumbs ~toc ~url ~uses_katex content subpages
+        ~breadcrumbs ~toc ~url ~uses_katex ~source_anchor content subpages
     else
+      let header =
+        items ~config ~resolve
+          (Doctree.PageTitle.render_title ?source_anchor p @ preamble)
+      in
       Html_page.make ~config ~header ~toc ~breadcrumbs ~url ~uses_katex content
         subpages
+
+  and source_page ~config sp =
+    let { Source_page.url; contents } = sp in
+    let resolve = Link.Current sp.url in
+    let title = url.Url.Path.name
+    and doc = Html_source.html_of_doc ~config ~resolve contents in
+    let breadcrumbs = Breadcrumbs.gen_breadcrumbs ~config ~url in
+    let header =
+      items ~config ~resolve (Doctree.PageTitle.render_src_title sp)
+    in
+    if Config.as_json config then
+      Html_fragment_json.make_src ~config ~url ~breadcrumbs [ doc ]
+    else Html_page.make_src ~breadcrumbs ~header ~config ~url title [ doc ]
+
+  let asset ~config { Asset.url; src } =
+    let filename = Link.Path.as_filename ~is_flat:(Config.flat config) url in
+    let content ppf =
+      let ic = open_in_bin (Fpath.to_string src) in
+      let len = 1024 in
+      let buf = Bytes.create len in
+      let rec loop () =
+        let read = input ic buf 0 len in
+        if read = len then (
+          Format.fprintf ppf "%s" (Bytes.to_string buf);
+          loop ())
+        else if len > 0 then
+          let buf = Bytes.sub buf 0 read in
+          Format.fprintf ppf "%s" (Bytes.to_string buf)
+      in
+      loop ();
+      close_in ic
+    in
+    { Odoc_document.Renderer.filename; content; children = [] }
 end
 
-let render ~config page = Page.page ~config page
+let render ~config = function
+  | Document.Page page -> [ Page.page ~config page ]
+  | Source_page src -> [ Page.source_page ~config src ]
+  | Asset asset -> [ Page.asset ~config asset ]
 
 let doc ~config ~xref_base_uri b =
   let resolve = Link.Base xref_base_uri in
