@@ -1,6 +1,6 @@
 val render :
   config:Config.t ->
-  Odoc_document.Types.Page.t ->
+  Odoc_document.Types.Document.t ->
   Odoc_document.Renderer.page list
 
 val doc :
