@@ -43,6 +43,8 @@ module Roff = struct
     | String of string
     | Vspace
     | Indent of int * t
+    | Align_line of string
+    | Table_cell of t
 
   let noop = Concat []
 
@@ -207,6 +209,13 @@ module Roff = struct
           | Macro (s, args) ->
               pp_macro ppf s "%s" args;
               newline_if ppf (not is_macro)
+          | Align_line s ->
+              Format.pp_print_string ppf (s ^ ".");
+              newline_if ppf (not is_macro)
+          | Table_cell c ->
+              Format.pp_print_text ppf "T{\n";
+              one ~indent ppf c;
+              Format.pp_print_text ppf "\nT}"
           | Indent (i, content) ->
               let indent = indent + i in
               one ~indent ppf content);
@@ -238,9 +247,7 @@ let strip l =
               { h with desc = Styled (sty, List.rev @@ loop [] content) }
             in
             loop (h :: acc) t
-        | Link (_, content)
-        | InternalLink (Resolved (_, content))
-        | InternalLink (Unresolved content) ->
+        | Link (_, content) | InternalLink { content; _ } ->
             let acc = loop acc content in
             loop acc t
         | Source code ->
@@ -298,12 +305,59 @@ and inline (l : Inline.t) =
       | Styled (sty, content) -> style sty (inline content) ++ inline rest
       | Link (href, content) ->
           env "UR" "UE" href (inline @@ strip content) ++ inline rest
-      | InternalLink (Resolved (_, content) | Unresolved content) ->
+      | InternalLink { content; _ } ->
           font "CI" (inline @@ strip content) ++ inline rest
       | Source content -> source_code content ++ inline rest
       | Math s -> math s ++ inline rest
       | Raw_markup t -> raw_markup t ++ inline rest)
 
+let table pp { Table.data; align } =
+  let sep = '\t' in
+  let alignment =
+    let alignment =
+      match align with
+      | align ->
+          List.map
+            (function
+              (* Since we are enclosing cells in text blocks, the alignment has
+                 no effect on the content of a sufficiently big cell, for some
+                 reason... (see the markup test in generators)
+
+                 One solution would be to use the [m] column specifier to apply
+                 a macro to the text blocks of the columns. Those macros would
+                 be [lj], [ce] or [rj], which define alignment. However, this
+                 breaks both the alignment for small table cells, and the
+                 largeness of columns. For the records, it woulb be:
+
+                 {[
+                   | Some `Left -> "lmlj"
+                   | Some `Center -> "cmce"
+                   | Some `Right -> "rmrj"
+                   | None -> "l"
+                  ]} *)
+              | Table.Left -> "l"
+              | Center -> "c"
+              | Right -> "r"
+              | Default -> "l")
+            align
+    in
+    Align_line (String.concat "" alignment)
+  in
+  env "TS" "TE" ""
+    (str "allbox;" ++ alignment
+    ++ List.fold_left
+         (fun acc row ->
+           acc ++ vspace
+           ++
+           match row with
+           | [] -> noop
+           | (h, _) :: t ->
+               List.fold_left
+                 (fun acc (x, _) -> acc ++ str "%c" sep ++ Table_cell (pp x))
+                 (Table_cell (pp h))
+                 t)
+         noop data)
+
 let rec block (l : Block.t) =
   match l with
   | [] -> noop
@@ -322,6 +376,7 @@ let rec block (l : Block.t) =
             indent 2 (bullet ++ sp ++ block b)
           in
           list ~sep:break (List.mapi f l) ++ continue rest
+      | Table t -> table block t ++ continue rest
       | Description _ ->
           let descrs, _, rest =
             Take.until l ~classify:(function
@@ -357,7 +412,7 @@ let next_heading, reset_heading =
   and reset () = heading_stack := [] in
   (next, reset)
 
-let heading ~nested { Heading.label = _; level; title } =
+let heading ~nested { Heading.label = _; level; title; source_anchor = _ } =
   let prefix =
     if level = 0 then noop
     else if level <= 3 then str "%s " (next_heading level)
@@ -431,7 +486,7 @@ let rec documentedSrc (l : DocumentedSrc.t) =
           let l = list ~sep:break (List.map f lines) in
           indent 2 (break ++ l) ++ break_if_nonempty rest ++ continue rest)
 
-and subpage { preamble = _; items; url = _ } =
+and subpage { preamble = _; items; url = _; _ } =
   let content = items in
   let surround body =
     if content = [] then sp else indent 2 (break ++ body) ++ break
@@ -450,7 +505,7 @@ and item ~nested (l : Item.t list) =
       | Heading h ->
           let h = heading ~nested h in
           vspace ++ h ++ vspace ++ item ~nested rest
-      | Declaration { attr = _; anchor = _; content; doc } ->
+      | Declaration { attr = _; anchor = _; source_anchor = _; content; doc } ->
           let decl = documentedSrc content in
           let doc =
             match doc with
@@ -459,8 +514,13 @@ and item ~nested (l : Item.t list) =
           in
           decl ++ doc ++ continue rest
       | Include
-          { attr = _; anchor = _; content = { summary; status; content }; doc }
-        ->
+          {
+            attr = _;
+            anchor = _;
+            source_anchor = _;
+            content = { summary; status; content };
+            doc;
+          } ->
           let d =
             if inline_subpage status then item ~nested content
             else
@@ -491,11 +551,15 @@ let page p =
 
 let rec subpage subp =
   let p = subp.Subpage.content in
-  if Link.should_inline p.url then [] else [ render p ]
+  if Link.should_inline p.url then [] else [ render_page p ]
 
-and render (p : Page.t) =
-  let p = Doctree.Labels.disambiguate_page p
+and render_page (p : Page.t) =
+  let p = Doctree.Labels.disambiguate_page ~enter_subpages:true p
   and children = Utils.flatmap ~f:subpage @@ Subpages.compute p in
   let content ppf = Format.fprintf ppf "%a@." Roff.pp (page p) in
   let filename = Link.as_filename p.url in
   { Renderer.filename; content; children }
+
+let render = function
+  | Document.Page page -> [ render_page page ]
+  | Source_page _ | Asset _ -> []
