@@ -18,6 +18,37 @@ module Identifier = struct
   and page = page_pv id
   (** @canonical Odoc_model.Paths.Identifier.Page.t *)
 
+  type source_dir_pv = [ container_page_pv | `SourceDir of source_dir * string ]
+  (** @canonical Odoc_model.Paths.Identifier.SourceDir.t_pv *)
+
+  and source_dir = source_dir_pv id
+  (** @canonical Odoc_model.Paths.Identifier.SourceDir.t *)
+
+  type source_page_pv = [ `SourcePage of source_dir * string ]
+  (** The second argument is the filename.
+
+     @canonical Odoc_model.Paths.Identifier.SourcePage.t_pv *)
+
+  type source_page = source_page_pv id
+  (** @canonical Odoc_model.Paths.Identifier.SourcePage.t *)
+
+  type asset_file_pv = [ `AssetFile of page * string ]
+  (** The second argument is the filename.
+
+    @canonical Odoc_model.Paths.Identifier.AssetFile.t_pv *)
+
+  type asset_file = asset_file_pv id
+  (** @canonical Odoc_model.Paths.Identifier.AssetFile.t *)
+
+  type source_location_pv =
+    [ `SourceLocationMod of source_page
+    | `SourceLocation of source_page * DefName.t
+    | `SourceLocationInternal of source_page * LocalName.t ]
+  (** @canonical Odoc_model.Paths.Identifier.SourceLocation.t *)
+
+  and source_location = source_location_pv id
+  (** @canonical Odoc_model.Paths.Identifier.SourceLocation.t_pv *)
+
   type odoc_id_pv = [ page_pv | `Root of container_page option * ModuleName.t ]
   (** @canonical Odoc_model.Paths.Identifier.OdocId.t_pv *)
 
@@ -50,13 +81,16 @@ module Identifier = struct
   and datatype = datatype_pv id
   (** @canonical Odoc_model.Paths.Identifier.DataType.t *)
 
-  type parent_pv = [ signature_pv | datatype_pv | class_signature_pv ]
-  (** @canonical Odoc_model.Paths.Identifier.Parent.t_pv *)
+  type field_parent_pv = [ signature_pv | datatype_pv ]
+  (** @canonical Odoc_model.Paths.Identifier.FieldParent.t_pv *)
 
-  and parent = parent_pv id
-  (** @canonical Odoc_model.Paths.Identifier.Parent.t *)
+  (* fragment_type_parent in identifiers is for record fields parent. It’s type
+     (for usual record fields) or [signature] for fields of inline records of
+     extension constructor. *)
+  and field_parent = field_parent_pv id
+  (** @canonical Odoc_model.Paths.Identifier.FieldParent.t *)
 
-  type label_parent_pv = [ parent_pv | page_pv ]
+  type label_parent_pv = [ field_parent_pv | page_pv | class_signature_pv ]
   (** @canonical Odoc_model.Paths.Identifier.LabelParent.t_pv *)
 
   and label_parent = label_parent_pv id
@@ -101,13 +135,13 @@ module Identifier = struct
   and type_ = type_pv id
   (** @canonical Odoc_model.Paths.Identifier.Type.t *)
 
-  type constructor_pv = [ `Constructor of type_ * ConstructorName.t ]
+  type constructor_pv = [ `Constructor of datatype * ConstructorName.t ]
   (** @canonical Odoc_model.Paths.Identifier.Constructor.t_pv *)
 
   and constructor = constructor_pv id
   (** @canonical Odoc_model.Paths.Identifier.Constructor.t *)
 
-  type field_pv = [ `Field of parent * FieldName.t ]
+  type field_pv = [ `Field of field_parent * FieldName.t ]
   (** @canonical Odoc_model.Paths.Identifier.Field.t_pv *)
 
   and field = field_pv id
@@ -116,9 +150,16 @@ module Identifier = struct
   type extension_pv = [ `Extension of signature * ExtensionName.t ]
   (** @canonical Odoc_model.Paths.Identifier.Extension.t_pv *)
 
+  type extension_decl_pv =
+    [ `ExtensionDecl of signature * ExtensionName.t * ExtensionName.t ]
+  (** @canonical Odoc_model.Paths.Identifier.ExtensionDecl.t_pv *)
+
   and extension = extension_pv id
   (** @canonical Odoc_model.Paths.Identifier.Extension.t *)
 
+  and extension_decl = extension_decl_pv id
+  (** @canonical Odoc_model.Paths.Identifier.ExtensionDecl.t *)
+
   type exception_pv =
     [ `Exception of signature * ExceptionName.t
     | `CoreException of ExceptionName.t ]
@@ -164,11 +205,11 @@ module Identifier = struct
   and label = label_pv id
   (** @canonical Odoc_model.Paths.Identifier.Label.t *)
 
-  type any_pv =
+  type non_src_pv =
     [ signature_pv
     | class_signature_pv
     | datatype_pv
-    | parent_pv
+    | field_parent_pv
     | label_parent_pv
     | module_pv
     | functor_parameter_pv
@@ -178,6 +219,7 @@ module Identifier = struct
     | constructor_pv
     | field_pv
     | extension_pv
+    | extension_decl_pv
     | exception_pv
     | value_pv
     | class_pv
@@ -186,6 +228,17 @@ module Identifier = struct
     | instance_variable_pv
     | label_pv
     | page_pv ]
+  (** @canonical Odoc_model.Paths.Identifier.NonSrc.t_pv *)
+
+  and non_src = non_src_pv id
+  (** @canonical Odoc_model.Paths.Identifier.NonSrc.t *)
+
+  type any_pv =
+    [ non_src_pv
+    | source_page_pv
+    | source_dir_pv
+    | source_location_pv
+    | asset_file_pv ]
   (** @canonical Odoc_model.Paths.Identifier.t_pv *)
 
   and any = any_pv id
@@ -206,6 +259,16 @@ module Identifier = struct
   and path_type = path_type_pv id
   (** @canonical Odoc_model.Paths.Identifier.Path.Type.t *)
 
+  type path_datatype_pv = type_pv
+  (** @canonical Odoc_model.Paths.Identifier.Path.DataType.t_pv *)
+
+  and path_datatype = path_datatype_pv id
+  (** @canonical Odoc_model.Paths.Identifier.Path.DataType.t *)
+
+  type path_constructor = constructor
+
+  type path_value = value
+
   type path_class_type_pv = [ class_pv | class_type_pv ]
   (** @canonical Odoc_model.Paths.Identifier.Path.ClassType.t_pv *)
 
@@ -213,7 +276,12 @@ module Identifier = struct
   (** @canonical Odoc_model.Paths.Identifier.Path.ClassType.t *)
 
   type path_any =
-    [ path_module_pv | module_type_pv | path_type_pv | path_class_type_pv ] id
+    [ path_module_pv
+    | module_type_pv
+    | path_type_pv
+    | path_class_type_pv
+    | value_pv ]
+    id
   (** @canonical Odoc_model.Paths.Identifier.Path.t *)
 
   type fragment_module = path_module
@@ -233,6 +301,8 @@ module Identifier = struct
 
   type reference_extension = [ extension_pv | exception_pv ] id
 
+  type reference_extension_decl = extension_decl
+
   type reference_exception = exception_
 
   type reference_value = value
@@ -272,6 +342,19 @@ module rec Path : sig
     | `Dot of module_ * string ]
   (** @canonical Odoc_model.Paths.Path.Type.t *)
 
+  type datatype =
+    [ `Resolved of Resolved_path.datatype
+    | `Identifier of Identifier.path_datatype * bool
+    | `Dot of module_ * string ]
+  (** @canonical Odoc_model.Paths.Path.DataType.t *)
+
+  type constructor =
+    [ `Resolved of Resolved_path.constructor | `Dot of datatype * string ]
+  (** @canonical Odoc_model.Paths.Path.Constructor.t *)
+
+  type value = [ `Resolved of Resolved_path.value | `Dot of module_ * string ]
+  (** @canonical Odoc_model.Paths.Path.Value.t *)
+
   type class_type =
     [ `Resolved of Resolved_path.class_type
     | `Identifier of Identifier.path_class_type * bool
@@ -318,6 +401,18 @@ and Resolved_path : sig
     | `ClassType of module_ * ClassTypeName.t ]
   (** @canonical Odoc_model.Paths.Path.Resolved.Type.t *)
 
+  type datatype =
+    [ `Identifier of Identifier.datatype
+    | `CanonicalDataType of datatype * Path.datatype
+    | `Type of module_ * TypeName.t ]
+  (** @canonical Odoc_model.Paths.Path.Resolved.DataType.t *)
+
+  type constructor = [ `Constructor of datatype * ConstructorName.t ]
+  (** @canonical Odoc_model.Paths.Path.Resolved.Constructor.t *)
+
+  type value = [ `Value of module_ * ValueName.t ]
+  (** @canonical Odoc_model.Paths.Path.Resolved.Value.t *)
+
   type class_type =
     [ `Identifier of Identifier.path_class_type
     | `Class of module_ * ClassName.t
@@ -338,10 +433,13 @@ and Resolved_path : sig
     | `SubstT of module_type * module_type
     | `OpaqueModuleType of module_type
     | `CanonicalType of type_ * Path.type_
+    | `CanonicalDataType of datatype * Path.datatype
     | `Type of module_ * TypeName.t
+    | `Constructor of datatype * ConstructorName.t
     | `Class of module_ * ClassName.t
     | `ClassType of module_ * ClassTypeName.t
     | `Class of module_ * ClassName.t
+    | `Value of module_ * ValueName.t
     | `ClassType of module_ * ClassTypeName.t ]
   (** @canonical Odoc_model.Paths.Path.Resolved.t *)
 end =
@@ -466,6 +564,7 @@ module rec Reference : sig
     | `TConstructor
     | `TField
     | `TExtension
+    | `TExtensionDecl
     | `TException
     | `TValue
     | `TClass
@@ -484,8 +583,7 @@ module rec Reference : sig
 
   type tag_datatype = [ `TUnknown | `TType ]
 
-  type tag_parent =
-    [ `TUnknown | `TModule | `TModuleType | `TClass | `TClassType | `TType ]
+  type tag_parent = [ `TUnknown | `TModule | `TModuleType | `TType ]
 
   type tag_label_parent =
     [ `TUnknown
@@ -521,16 +619,15 @@ module rec Reference : sig
     | `Type of signature * TypeName.t ]
   (** @canonical Odoc_model.Paths.Reference.DataType.t *)
 
-  and parent =
-    [ `Resolved of Resolved_reference.parent
+  (* Parent of fields and constructor. Can be either a type or [signature] *)
+  and fragment_type_parent =
+    [ `Resolved of Resolved_reference.field_parent
     | `Root of string * tag_parent
     | `Dot of label_parent * string
     | `Module of signature * ModuleName.t
     | `ModuleType of signature * ModuleTypeName.t
-    | `Class of signature * ClassName.t
-    | `ClassType of signature * ClassTypeName.t
     | `Type of signature * TypeName.t ]
-  (** @canonical Odoc_model.Paths.Reference.Parent.t *)
+  (** @canonical Odoc_model.Paths.Reference.FragmentTypeParent.t *)
 
   and label_parent =
     [ `Resolved of Resolved_reference.label_parent
@@ -570,7 +667,7 @@ module rec Reference : sig
     [ `Resolved of Resolved_reference.constructor
     | `Root of string * [ `TConstructor | `TExtension | `TException | `TUnknown ]
     | `Dot of label_parent * string
-    | `Constructor of datatype * ConstructorName.t
+    | `Constructor of fragment_type_parent * ConstructorName.t
     | `Extension of signature * ExtensionName.t
     | `Exception of signature * ExceptionName.t ]
   (** @canonical Odoc_model.Paths.Reference.Constructor.t *)
@@ -579,7 +676,7 @@ module rec Reference : sig
     [ `Resolved of Resolved_reference.field
     | `Root of string * [ `TField | `TUnknown ]
     | `Dot of label_parent * string
-    | `Field of parent * FieldName.t ]
+    | `Field of fragment_type_parent * FieldName.t ]
   (** @canonical Odoc_model.Paths.Reference.Field.t *)
 
   type extension =
@@ -590,6 +687,13 @@ module rec Reference : sig
     | `Exception of signature * ExceptionName.t ]
   (** @canonical Odoc_model.Paths.Reference.Extension.t *)
 
+  type extension_decl =
+    [ `Resolved of Resolved_reference.extension_decl
+    | `Root of string * [ `TExtension | `TException | `TUnknown ]
+    | `Dot of label_parent * string
+    | `ExtensionDecl of signature * ExtensionName.t ]
+  (** @canonical Odoc_model.Paths.Reference.ExtensionDecl.t *)
+
   type exception_ =
     [ `Resolved of Resolved_reference.exception_
     | `Root of string * [ `TException | `TUnknown ]
@@ -653,9 +757,10 @@ module rec Reference : sig
     | `Module of signature * ModuleName.t
     | `ModuleType of signature * ModuleTypeName.t
     | `Type of signature * TypeName.t
-    | `Constructor of datatype * ConstructorName.t
-    | `Field of parent * FieldName.t
+    | `Constructor of fragment_type_parent * ConstructorName.t
+    | `Field of fragment_type_parent * FieldName.t
     | `Extension of signature * ExtensionName.t
+    | `ExtensionDecl of signature * ExtensionName.t
     | `Exception of signature * ExceptionName.t
     | `Value of signature * ValueName.t
     | `Class of signature * ClassName.t
@@ -701,18 +806,18 @@ and Resolved_reference : sig
     | `ClassType of signature * ClassTypeName.t ]
   (** @canonical Odoc_model.Paths.Reference.Resolved.ClassSignature.t *)
 
-  (* parent is [ signature | class_signature ] *)
-  and parent =
-    [ `Identifier of Identifier.parent
+  (* fragment_type_parent in resolved references is for record fields parent.
+     It’s type (for usual record fields) or [signature] for fields of inline
+     records of extension constructor. *)
+  and field_parent =
+    [ `Identifier of Identifier.field_parent
     | `Alias of Resolved_path.module_ * module_
     | `AliasModuleType of Resolved_path.module_type * module_type
     | `Module of signature * ModuleName.t
     | `Hidden of module_
     | `ModuleType of signature * ModuleTypeName.t
-    | `Class of signature * ClassName.t
-    | `ClassType of signature * ClassTypeName.t
     | `Type of signature * TypeName.t ]
-  (** @canonical Odoc_model.Paths.Reference.Resolved.Parent.t *)
+  (** @canonical Odoc_model.Paths.Reference.Resolved.FragmentTypeParent.t *)
 
   (* The only difference between parent and label_parent
      is that the Identifier allows more types *)
@@ -750,7 +855,7 @@ and Resolved_reference : sig
 
   type field =
     [ `Identifier of Identifier.reference_field
-    | `Field of parent * FieldName.t ]
+    | `Field of field_parent * FieldName.t ]
   (** @canonical Odoc_model.Paths.Reference.Resolved.Field.t *)
 
   type extension =
@@ -759,6 +864,16 @@ and Resolved_reference : sig
     | `Exception of signature * ExceptionName.t ]
   (** @canonical Odoc_model.Paths.Reference.Resolved.Extension.t *)
 
+  type extension_decl =
+    [ `Identifier of Identifier.reference_extension_decl
+    | `ExtensionDecl of
+      signature
+      * ExtensionName.t
+        (* The extension_name used in the url.
+           It is the extension_name of the first constructor of the extension (there is always at least 1). *)
+      * ExtensionName.t (* displayed *) ]
+  (** @canonical Odoc_model.Paths.Reference.Resolved.Extension.t *)
+
   type exception_ =
     [ `Identifier of Identifier.reference_exception
     | `Exception of signature * ExceptionName.t ]
@@ -807,8 +922,9 @@ and Resolved_reference : sig
     | `ModuleType of signature * ModuleTypeName.t
     | `Type of signature * TypeName.t
     | `Constructor of datatype * ConstructorName.t
-    | `Field of parent * FieldName.t
+    | `Field of field_parent * FieldName.t
     | `Extension of signature * ExtensionName.t
+    | `ExtensionDecl of signature * ExtensionName.t * ExtensionName.t
     | `Exception of signature * ExceptionName.t
     | `Value of signature * ValueName.t
     | `Class of signature * ClassName.t
