@@ -29,6 +29,40 @@ let convert_directory ?(create = false) () : Fs.Directory.t Arg.conv =
   let odoc_dir_printer fmt dir = dir_printer fmt (Fs.Directory.to_string dir) in
   (odoc_dir_parser, odoc_dir_printer)
 
+(** On top of the conversion 'file' that checks that the passed file exists. *)
+let convert_fpath =
+  let parse inp =
+    match Arg.(conv_parser file) inp with
+    | Ok s -> Result.Ok (Fs.File.of_string s)
+    | Error _ as e -> e
+  and print = Fpath.pp in
+  Arg.conv (parse, print)
+
+let convert_src_fpath =
+  let parse inp =
+    match Arg.(conv_parser file) inp with
+    | Ok s -> Result.Ok (Html_page.Source.File (Fs.File.of_string s))
+    | Error _ as e -> e
+  and print = Html_page.Source.pp in
+  Arg.conv (parse, print)
+
+let convert_src_dir =
+  let parse inp =
+    match Arg.(conv_parser dir) inp with
+    | Ok s -> Result.Ok (Html_page.Source.Root (Fs.File.of_string s))
+    | Error _ as e -> e
+  and print = Html_page.Source.pp in
+  Arg.conv (parse, print)
+
+(** On top of the conversion 'string', split into segs. *)
+let convert_source_name =
+  let parse inp =
+    match Arg.(conv_parser string) inp with
+    | Ok s -> Result.Ok (s |> Fs.File.of_string |> Fs.File.segs)
+    | Error _ as e -> e
+  and print ppf x = Format.fprintf ppf "%s" (String.concat ~sep:"/" x) in
+  Arg.conv (parse, print)
+
 let handle_error = function
   | Result.Ok () -> ()
   | Error (`Cli_error msg) ->
@@ -149,7 +183,8 @@ end = struct
         Fs.File.(set_ext ".odoc" output)
 
   let compile hidden directories resolve_fwd_refs dst package_opt
-      parent_name_opt open_modules children input warnings_options =
+      parent_name_opt open_modules children input warnings_options
+      source_parent_file source_name cmt_filename_opt =
     let open Or_error in
     let resolver =
       Resolver.create ~important_digests:(not resolve_fwd_refs) ~directories
@@ -167,10 +202,25 @@ end = struct
             (`Cli_error
               "Either --package or --parent should be specified, not both")
     in
+    let source =
+      match (source_parent_file, source_name) with
+      | Some parent, Some name -> Ok (Some (parent, name))
+      | Some _, None | None, Some _ ->
+          Error
+            (`Cli_error
+              "--source-parent-file and --source-name must be passed at the \
+               same time.")
+      | None, None -> Ok None
+    in
+    (if Fs.File.get_ext input = ".cmt" && cmt_filename_opt <> None then
+       Error (`Cli_error "--cmt is redundant if the input is a cmt file")
+     else Ok ())
+    >>= fun () ->
     parent_cli_spec >>= fun parent_cli_spec ->
+    source >>= fun source ->
     Fs.Directory.mkdir_p (Fs.File.dirname output);
     Compile.compile ~resolver ~parent_cli_spec ~hidden ~children ~output
-      ~warnings_options input
+      ~warnings_options ~source ~cmt_filename_opt input
 
   let input =
     let doc = "Input $(i,.cmti), $(i,.cmt), $(i,.cmi) or $(i,.mld) file." in
@@ -189,12 +239,39 @@ end = struct
   let children =
     let doc =
       "Specify the $(i,.odoc) file as a child. Can be used multiple times. \
-       Only applies to mld files"
+       Only applies to mld files."
     in
     let default = [] in
     Arg.(
       value & opt_all string default & info ~docv:"CHILD" ~doc [ "c"; "child" ])
 
+  let source_parent_file =
+    let doc =
+      ".odoc file of the parent of the page containing the source code for \
+       this compilation unit."
+    in
+    Arg.(
+      value
+      & opt (some convert_fpath) None
+      & info [ "source-parent-file" ] ~doc ~docv:"PARENT.odoc")
+
+  let source_name =
+    let doc =
+      "The basename of the source file. This is used to place the source file \
+       within the source_parent."
+    in
+    Arg.(
+      value
+      & opt (some convert_source_name) None
+      & info [ "source-name" ] ~doc ~docv:"NAME")
+
+  let source_cmt =
+    let doc =
+      "The .cmt file to use for source code related operations, such as source \
+       code rendering."
+    in
+    Arg.(value & opt (some file) None & info [ "cmt" ] ~doc ~docv:"CMT")
+
   let cmd =
     let package_opt =
       let doc =
@@ -220,7 +297,7 @@ end = struct
       const handle_error
       $ (const compile $ hidden $ odoc_file_directories $ resolve_fwd_refs $ dst
        $ package_opt $ parent_opt $ open_modules $ children $ input
-       $ warnings_options))
+       $ warnings_options $ source_parent_file $ source_name $ source_cmt))
 
   let info ~docs =
     let man =
@@ -239,6 +316,135 @@ end = struct
     Term.info "compile" ~docs ~doc ~man
 end
 
+module Source_tree = struct
+  let has_src_prefix input =
+    input |> Fs.File.basename |> Fs.File.to_string
+    |> Astring.String.is_prefix ~affix:"src-"
+
+  let output_file ~output ~input =
+    match output with
+    | Some output -> output
+    | None ->
+        let output =
+          if not (has_src_prefix input) then
+            let directory = Fs.File.dirname input in
+            let name = input |> Fs.File.basename |> Fs.File.to_string in
+            let name = "src-" ^ name in
+            Fs.File.create ~directory ~name
+          else input
+        in
+        Fs.File.(set_ext ".odoc" output)
+
+  let compile_source_tree directories output parent input warnings_options =
+    let output = output_file ~output ~input in
+    let resolver =
+      Resolver.create ~important_digests:true ~directories ~open_modules:[]
+    in
+    Source_tree.compile ~resolver ~parent ~output ~warnings_options input
+
+  let arg_page_output =
+    let open Or_error in
+    let parse inp =
+      match Arg.(conv_parser string) inp with
+      | Ok s ->
+          let f = Fs.File.of_string s in
+          if not (Fs.File.has_ext ".odoc" f) then
+            Error (`Msg "Output file must have '.odoc' extension.")
+          else if not (has_src_prefix f) then
+            Error (`Msg "Output file must be prefixed with 'src-'.")
+          else Ok f
+      | Error _ as e -> e
+    and print = Fpath.pp in
+    Arg.conv (parse, print)
+
+  let cmd =
+    let parent =
+      let doc = "Parent page or subpage." in
+      Arg.(
+        required
+        & opt (some string) None
+        & info ~docs ~docv:"PARENT" ~doc [ "parent" ])
+    in
+    let dst =
+      let doc =
+        "Output file path. Non-existing intermediate directories are created. \
+         The basename must start with the prefix 'src-' and extension '.odoc'."
+      in
+      Arg.(
+        value
+        & opt (some arg_page_output) None
+        & info ~docs ~docv:"PATH" ~doc [ "o" ])
+    in
+    let input =
+      let doc = "Input text file containing a line-separated list of paths." in
+      Arg.(
+        required & pos 0 (some convert_fpath) None & info ~doc ~docv:"FILE" [])
+    in
+    Term.(
+      const handle_error
+      $ (const compile_source_tree $ odoc_file_directories $ dst $ parent
+       $ input $ warnings_options))
+
+  let info ~docs =
+    let doc =
+      "Compile a source tree into a page. Expect a text file containing the \
+       relative paths to every source files in the source tree. The paths \
+       should be the same as the one passed to $(i,odoc compile \
+       --source-name)."
+    in
+    Term.info "source-tree" ~docs ~doc
+end
+
+module Indexing = struct
+  let output_file ~dst =
+    match dst with
+    | Some file -> Fs.File.of_string file
+    | None -> Fs.File.of_string "index.json"
+
+  let index dst warnings_options inputs_in_file inputs =
+    let output = output_file ~dst in
+    match (inputs_in_file, inputs) with
+    | [], [] ->
+        Result.Error
+          (`Msg
+            "At least one of --file-list or an .odocl file must be passed to \
+             odoc compile-index")
+    | _ -> Indexing.compile ~output ~warnings_options inputs_in_file inputs
+
+  let cmd =
+    let dst =
+      let doc =
+        "Output file path. Non-existing intermediate directories are created. \
+         Defaults to index.json"
+      in
+      Arg.(
+        value & opt (some string) None & info ~docs ~docv:"PATH" ~doc [ "o" ])
+    in
+    let inputs_in_file =
+      let doc =
+        "Input text file containing a line-separated list of paths to .odocl \
+         files to index."
+      in
+      Arg.(
+        value & opt_all convert_fpath []
+        & info ~doc ~docv:"FILE" [ "file-list" ])
+    in
+    let inputs =
+      let doc = ".odocl file to index" in
+      Arg.(value & pos_all convert_fpath [] & info ~doc ~docv:"FILE" [])
+    in
+    Term.(
+      const handle_error
+      $ (const index $ dst $ warnings_options $ inputs_in_file $ inputs))
+
+  let info ~docs =
+    let doc =
+      "Generate an index of all identified entries in the .odocl files found \
+       in the given directories."
+    in
+    Term.info "compile-index" ~docs ~doc
+end
+
 module Support_files_command = struct
   let support_files without_theme output_dir =
     Support_files.write ~without_theme output_dir
@@ -387,11 +593,11 @@ end = struct
   let process ~docs = Process.(cmd, info ~docs)
 
   module Generate = struct
-    let generate extra _hidden output_dir syntax extra_suffix input_file =
+    let generate extra _hidden output_dir syntax extra_suffix input_file
+        warnings_options =
       let file = Fs.File.of_string input_file in
-
-      Rendering.generate_odoc ~renderer:R.renderer ~syntax ~output:output_dir
-        ~extra_suffix extra file
+      Rendering.generate_odoc ~renderer:R.renderer ~warnings_options ~syntax
+        ~output:output_dir ~extra_suffix extra file
 
     let cmd =
       let syntax =
@@ -405,7 +611,7 @@ end = struct
       Term.(
         const handle_error
         $ (const generate $ R.extra_args $ hidden $ dst ~create:true () $ syntax
-         $ extra_suffix $ input_odocl))
+         $ extra_suffix $ input_odocl $ warnings_options))
 
     let info ~docs =
       let doc =
@@ -478,9 +684,7 @@ end = struct
 end
 
 module Odoc_html_args = struct
-  type args = Odoc_html.Config.t
-
-  let renderer = Html_page.renderer
+  include Html_page
 
   let semantic_uris =
     let doc = "Generate pretty (semantic) links." in
@@ -497,59 +701,104 @@ module Odoc_html_args = struct
     let doc = "Format the output HTML files with indentation." in
     Arg.(value & flag (info ~doc [ "indent" ]))
 
-  (* Very basic validation and normalization for URI paths. *)
-  let convert_uri : Odoc_html.Types.uri Arg.conv =
-    let parser str =
-      if String.length str = 0 then `Error "invalid URI"
-      else
-        (* The URI is absolute if it starts with a scheme or with '/'. *)
-        let is_absolute =
-          List.exists [ "http"; "https"; "file"; "data"; "ftp" ]
-            ~f:(fun scheme ->
-              Astring.String.is_prefix ~affix:(scheme ^ ":") str)
-          || str.[0] = '/'
-        in
-        let last_char = str.[String.length str - 1] in
-        let str =
-          if last_char <> '/' then str
-          else String.sub str ~pos:0 ~len:(String.length str - 1)
-        in
-        let conv_rel rel =
-          let l = Astring.String.cuts ~sep:"/" rel in
-          List.fold_left
-            ~f:(fun acc seg ->
-              Some
+  module Uri = struct
+    (* Very basic validation and normalization for URI paths. *)
+
+    open Odoc_html.Types
+
+    let is_absolute str =
+      List.exists [ "http"; "https"; "file"; "data"; "ftp" ] ~f:(fun scheme ->
+          Astring.String.is_prefix ~affix:(scheme ^ ":") str)
+      || str.[0] = '/'
+
+    let conv_rel_dir rel =
+      let l = Astring.String.cuts ~sep:"/" rel in
+      List.fold_left
+        ~f:(fun acc seg ->
+          Some Odoc_document.Url.Path.{ kind = `Page; parent = acc; name = seg })
+        l ~init:None
+
+    let convert_dir : uri Arg.conv =
+      let parser str =
+        if String.length str = 0 then `Error "invalid URI"
+        else
+          (* The URI is absolute if it starts with a scheme or with '/'. *)
+          let last_char = str.[String.length str - 1] in
+          let str =
+            if last_char <> '/' then str
+            else String.sub str ~pos:0 ~len:(String.length str - 1)
+          in
+          `Ok
+            (if is_absolute str then (Absolute str : uri)
+             else
+               Relative
+                 (let u = conv_rel_dir str in
+                  match u with
+                  | None -> None
+                  | Some u -> Some { u with kind = `Page }))
+      in
+      let printer ppf = function
+        | (Absolute uri : uri) -> Format.pp_print_string ppf uri
+        | Relative _uri -> Format.pp_print_string ppf ""
+      in
+      (parser, printer)
+
+    let convert_file_uri : Odoc_html.Types.file_uri Arg.conv =
+      let parser str =
+        if String.length str = 0 then `Error "invalid URI"
+        else
+          let conv_rel_file rel =
+            match Astring.String.cut ~rev:true ~sep:"/" rel with
+            | Some (before, after) ->
+                let base = conv_rel_dir before in
                 Odoc_document.Url.Path.
-                  { kind = `Page; parent = acc; name = seg })
-            l ~init:None
-        in
-        `Ok
-          Odoc_html.Types.(
-            if is_absolute then Absolute str else Relative (conv_rel str))
-    in
-    let printer ppf = function
-      | Odoc_html.Types.Absolute uri -> Format.pp_print_string ppf uri
-      | Odoc_html.Types.Relative _uri -> Format.pp_print_string ppf ""
-    in
-    (parser, printer)
+                  { kind = `File; parent = base; name = after }
+            | None ->
+                Odoc_document.Url.Path.
+                  { kind = `File; parent = None; name = rel }
+          in
+          `Ok
+            (if is_absolute str then (Absolute str : file_uri)
+             else Relative (conv_rel_file str))
+      in
+      let printer ppf = function
+        | Odoc_html.Types.Absolute uri -> Format.pp_print_string ppf uri
+        | Odoc_html.Types.Relative _uri -> Format.pp_print_string ppf ""
+      in
+      (parser, printer)
+  end
 
   let theme_uri =
     let doc =
       "Where to look for theme files (e.g. `URI/odoc.css'). Relative URIs are \
        resolved using `--output-dir' as a target."
     in
-    let default = Odoc_html.Types.Relative None in
+    let default : Odoc_html.Types.uri = Odoc_html.Types.Relative None in
     Arg.(
-      value & opt convert_uri default & info ~docv:"URI" ~doc [ "theme-uri" ])
+      value
+      & opt Uri.convert_dir default
+      & info ~docv:"URI" ~doc [ "theme-uri" ])
 
   let support_uri =
     let doc =
       "Where to look for support files (e.g. `URI/highlite.pack.js'). Relative \
        URIs are resolved using `--output-dir' as a target."
     in
-    let default = Odoc_html.Types.Relative None in
+    let default : Odoc_html.Types.uri = Odoc_html.Types.Relative None in
     Arg.(
-      value & opt convert_uri default & info ~docv:"URI" ~doc [ "support-uri" ])
+      value
+      & opt Uri.convert_dir default
+      & info ~docv:"URI" ~doc [ "support-uri" ])
+
+  let search_uri =
+    let doc =
+      "Where to look for search scripts. Relative URIs are resolved using \
+       `--output-dir' as a target."
+    in
+    Arg.(
+      value
+      & opt_all Uri.convert_file_uri []
+      & info ~docv:"URI" ~doc [ "search-uri" ])
 
   let flat =
     let doc =
@@ -569,16 +818,58 @@ module Odoc_html_args = struct
     in
     Arg.(value & flag & info ~doc [ "as-json" ])
 
+  let source_file =
+    let doc =
+      "Source code for the compilation unit. It must have been compiled with \
+       --source-parent passed."
+    in
+    Arg.(
+      value
+      & opt (some convert_src_fpath) None
+      & info [ "source" ] ~doc ~docv:"file.ml")
+
+  let assets =
+    let doc =
+      "Assets files. These must match the assets listed as children during the \
+       compile phase."
+    in
+    Arg.(
+      value & opt_all convert_fpath [] & info [ "asset" ] ~doc ~docv:"file.ext")
+
+  let source_root =
+    let doc =
+      "Source code root for the compilation unit. Used to find the source file \
+       from the value of --source-name it was compiled with. Incompatible with \
+       --source-file."
+    in
+    Arg.(
+      value
+      & opt (some convert_src_dir) None
+      & info [ "source-root" ] ~doc ~docv:"dir")
+
   let extra_args =
-    let config semantic_uris closed_details indent theme_uri support_uri flat
-        as_json =
+    let config semantic_uris closed_details indent theme_uri support_uri
+        search_uris flat as_json source_file assets source_root =
       let open_details = not closed_details in
-      Odoc_html.Config.v ~theme_uri ~support_uri ~semantic_uris ~indent ~flat
-        ~open_details ~as_json ()
+      let source =
+        match (source_root, source_file) with
+        | Some x, None -> Some x
+        | None, Some x -> Some x
+        | None, None -> None
+        | Some _, Some _ ->
+            Printf.eprintf "ERROR: Can't use both source and source-root\n%!";
+            exit 1
+      in
+      let html_config =
+        Odoc_html.Config.v ~theme_uri ~support_uri ~search_uris ~semantic_uris
+          ~indent ~flat ~open_details ~as_json ()
+      in
+      { Html_page.html_config; source; assets }
     in
     Term.(
       const config $ semantic_uris $ closed_details $ indent $ theme_uri
-      $ support_uri $ flat $ as_json)
+      $ support_uri $ search_uri $ flat $ as_json $ source_file $ assets
+      $ source_root)
 end
 
 module Odoc_html = Make_renderer (Odoc_html_args)
@@ -685,8 +976,10 @@ end)
 
 module Depends = struct
   module Compile = struct
-    let list_dependencies input_file =
-      let deps = Depends.for_compile_step (Fs.File.of_string input_file) in
+    let list_dependencies input_files =
+      let deps =
+        Depends.for_compile_step (List.map ~f:Fs.File.of_string input_files)
+      in
       List.iter
         ~f:(fun t ->
           Printf.printf "%s %s\n" (Depends.Compile.name t)
@@ -696,11 +989,8 @@ module Depends = struct
 
     let cmd =
       let input =
-        let doc = "Input file" in
-        Arg.(
-          required
-          & pos 0 (some file) None
-          & info ~doc ~docv:"file.cm{i,t,ti}" [])
+        let doc = "Input files" in
+        Arg.(non_empty & pos_all file [] & info ~doc ~docv:"file.cm{i,t,ti}" [])
       in
       Term.(const list_dependencies $ input)
 
@@ -857,6 +1147,8 @@ let () =
       Odoc_link.(cmd, info ~docs:section_pipeline);
       Odoc_html.generate ~docs:section_pipeline;
       Support_files_command.(cmd, info ~docs:section_pipeline);
+      Source_tree.(cmd, info ~docs:section_pipeline);
+      Indexing.(cmd, info ~docs:section_pipeline);
       Odoc_manpage.generate ~docs:section_generators;
       Odoc_latex.generate ~docs:section_generators;
       Odoc_html_url.(cmd, info ~docs:section_support);
@@ -891,7 +1183,7 @@ let () =
       List.map ~f:(fun s -> `S s) main_page_sections
     in
     ( Term.(const print_default $ const ()),
-      Term.info ~man ~version:"2.2.1" "odoc" )
+      Term.info ~man ~version:"2.4.1" "odoc" )
   in
   match Term.eval_choice ~err:Format.err_formatter default subcommands with
   | `Error _ ->
