@@ -564,11 +564,14 @@ module LangUtils = struct
             | `Module (p, m) -> Format.fprintf ppf "%a.%s" resolved_path (cast p) (Odoc_model.Names.ModuleName.to_string m)
             | `ModuleType (p, mt) -> Format.fprintf ppf "%a.%s" resolved_path (cast p) (Odoc_model.Names.ModuleTypeName.to_string mt)
             | `Type (p, t) -> Format.fprintf ppf "%a.%s" resolved_path (cast p) (Odoc_model.Names.TypeName.to_string t)
+            | `Value (p, t) -> Format.fprintf ppf "%a.%s" resolved_path (cast p) (Odoc_model.Names.ValueName.to_string t)
+            | `Constructor (p, t) -> Format.fprintf ppf "%a.%s" resolved_path (cast p) (Odoc_model.Names.ConstructorName.to_string t)
             | `OpaqueModule m -> Format.fprintf ppf "opaquemodule(%a)" resolved_path (cast m)
             | `OpaqueModuleType m -> Format.fprintf ppf "opaquemoduletype(%a)" resolved_path (cast m)
             | `SubstT (_, _)
             | `CanonicalModuleType (_, _)
             | `CanonicalType (_, _)
+            | `CanonicalDataType (_, _)
             | `Class (_, _)
             | `ClassType (_, _)
             | `Hidden _
@@ -615,6 +618,8 @@ let my_compilation_unit id (s : Odoc_model.Lang.Signature.t) =
     ; expansion = None
     ; linked = false
     ; canonical = None
+    ; source_info = None
+    ; shape_info = None
 }
 
 let mkresolver () =
@@ -638,7 +643,7 @@ let handle_warnings ww =
 
 let resolve unit =
   let resolver = mkresolver () in
-  let resolve_env = Odoc_odoc.Resolver.build_env_for_unit resolver ~linking:true unit in
+  let resolve_env = Odoc_odoc.Resolver.build_compile_env_for_unit resolver unit in
   Odoc_xref2.Compile.compile ~filename:"<test>" resolve_env unit
   |> handle_warnings
 
